(* Abstract Syntax Tree, printing functions *)

(* Authors: Kundan Guha, Yuanxia Lee *)

(* Variable names self-evaluate, arrays arbitrarily contain names or bits *)
type expr =
    Name of string
  | Bit of bool
  | Arr of expr list
  | Index of string * int

type prim = And | Or | Not (*| Dff*)

(* Type (prim or custom name), variable name *)
type vdecl =
    Pdec of prim * string
  | Gdec of string * string
  | Bdec of string
  | Barr of string * int

type iden =
    Symb of string
  | Deref of string * int

(* For semantic checking useful to know number of outputs expected *)
type gate_def = {
    name: string;
    inp: string list;
    body: stmt list;
    out: expr;
} and stmt =
    Assign of iden * expr
  | Gdef of gate_def
  | Vdec of vdecl
  | Vset of string * expr list

(* Variables split up into declaration and assignment even if not written that way *)


(*Begin attempt at pretty printing functions*)
let rec string_of_expr = function
    Name(s) -> s
  | Bit(true) -> "true"
  | Bit(false) -> "false" 
  | Arr(exprs) -> "[" ^ String.concat "" (List.map string_of_expr exprs) ^ "]"
  | Index(s,i) -> s ^ " " ^ string_of_int i 

let string_of_prim = function
    And -> "AND"
  | Not -> "NOT"
  | Or -> "OR"
  (*| Dff -> "DFF" *)

let string_of_vdecl = function
    Pdec(p,s) -> string_of_prim p ^ " " ^ s
  | Gdec(s1,s2) -> s1 ^ " " ^ s2
  | Bdec(s) -> s 
  | Barr(s,i) -> s ^ " " ^ string_of_int i 

let string_of_iden = function
    Symb s -> ("Symb " ^ s)
  | Deref (s, i) -> ("Deref " ^ s ^ "[" ^ string_of_int i ^ "]")

let rec string_of_inp = function
    [] -> ""
  | h::t -> h ^ (string_of_inp t)

let rec string_of_stmt = function
    Assign(s,e) -> (string_of_iden s) ^ "=" ^ string_of_expr e
  | Gdef(g) -> string_of_gate_def g
  | Vdec(v) -> string_of_vdecl v
  | Vset(s,e) -> s ^ " " ^ String.concat "" (List.map string_of_expr e)
and string_of_gate_def gdef = 
    (gdef.name) ^ "\n" ^ 
    (string_of_inp gdef.inp)^ "\n" ^ 
    (String.concat "\n" (List.map string_of_stmt gdef.body))^ "\n"^
    (string_of_expr gdef.out) ^ "\n"
