(*===----------------------------------------------------------------------===
 * Code Generation
 *===----------------------------------------------------------------------===*)

(* Authors: Kundan Guha, Hannah Pierce-Hoffman, Sarah Walker*)

module L = Llvm
module A = Netlist

module StringMap = Map.Make(String)

let translate (A.Program (nam, inputs, prims, outputs)) =

    let context = L.global_context () in
    let the_module = L.create_module context "Logisimple" 
    and void = L.void_type context
    and i8_t = L.i8_type context
    and i1_t = L.i1_type context in
    let i8p_t = L.pointer_type i8_t
    and zero = L.const_int i8_t 0
    and one = L.const_int i8_t 1 in
    let extargs = [ "Inp"; "Out" ] in

    let ftype = L.function_type void (Array.make (List.length extargs) i8p_t) in
    let mainf = L.define_function nam ftype the_module in
    let builder = L.builder_at_end context (L.entry_block mainf) in
    let formals = 
        let make_formals l n p = 
            L.set_value_name n p;
            let local = L.build_alloca i8p_t n builder in
                ignore (L.build_store p local builder);
                local :: l in
        List.rev (List.fold_left2 make_formals [] extargs (Array.to_list (L.params mainf)))
    in
    let ext_vars =
        let build_input nam count map =
            let arg = L.build_alloca i8_t nam builder
            and arrp = L.build_load (List.hd formals) "" builder in
            let arr = L.build_in_bounds_gep arrp [| L.const_int i8_t count |] "input" builder in
            let arr1 = L.build_load arr "" builder in
            ignore (L.build_store arr1 arg builder);
            StringMap.add nam arg map in
        let rec build_inputs count l map =
            match l with
              [] -> map
            | h :: t -> build_inputs (count + 1) t (build_input h count map) in
        build_inputs 0 inputs StringMap.empty in
    (*let rec print_keys l = match l with
      [] -> print_endline "contained" 
    | (x, y) :: t -> print_bytes (x ^ " "); print_keys t in
    ignore (print_keys (StringMap.bindings vars));*)
    let lookup n vars = try StringMap.find n vars
                   with Not_found -> print_endline ("Key: " ^ n ^ " not found.");
                   raise Not_found in
    let codegen_arg a vars builder =
        match a with
          A.Name n -> L.build_load (lookup n vars) n builder
        | A.Bool b -> if b then one else zero
    in
    let expr builder vars (A.Expr (out, gate)) = 
        let store_output = L.build_alloca i8_t out builder in
        (match gate with
          A.Prim (op, in1, in2) -> 
            let pt1 = codegen_arg in1 vars builder and
                pt2 = codegen_arg in2 vars builder in
                let and_result = 
                    (match op with
                        A.And -> L.build_and
                        | A.Or -> L.build_or
                    ) pt1 pt2 "tmp" builder in
                    ignore (L.build_store and_result store_output builder);
        | A.Neg in1 -> 
            let pt5 = codegen_arg in1 vars builder in

            let not_result = L.build_not pt5 "tmp" builder in let not_result =
                L.build_and not_result one "tmp" builder in
            ignore (L.build_store not_result store_output builder);
        );
        StringMap.add out store_output vars
    in
    let nlist vars (e: A.stmt) = expr builder vars e
    in
    let all_vars = List.fold_left nlist ext_vars prims in
    let output =
        let build_output nam count map =
            let arg = L.build_load (lookup nam map) nam builder in
            let arrp = L.build_load (List.hd (List.tl formals)) "" builder in
            let arr = L.build_in_bounds_gep arrp [| L.const_int i8_t count |] "output" builder in
            ignore (L.build_store arg arr builder); map in
        let rec build_outputs count l map =
            match l with
              [] -> ()
            | h :: t -> build_outputs (count + 1) t (build_output h count map) in
        build_outputs 0 outputs all_vars in
    ignore (L.build_ret_void builder);
    the_module
