(* Authors: Kundan Guha, Yuanxia Lee *)

open Ast

module StringMap = Map.Make(String)
module N = Netlist

let check {name; inp; body; out} =
    let report_duplicate exceptf l =
        let rec helper = function
          n1 :: n2 :: _ when n1 = n2 -> raise (Failure (exceptf n1))
        | _ :: t -> helper t
        | [] -> () in
    helper (List.sort compare l) in

    let add_inputs input map = List.fold_left 
        (fun m n -> StringMap.add n (Bdec n) m) map input in

    let merge_scopes local global =
        let mf k lv gv = match lv, gv with
          v, None -> v
        | None, v -> v
        | (Some v), (Some w) -> lv in
        StringMap.merge mf local global in

    let build_output_list out =
        let name_to_str e =
            match e with
              Name s -> s
            | _ -> raise (Failure ("Non-identifier in output list")) in
        match out with
          Name s -> [s]
        | Arr el -> List.map name_to_str el
        | _ -> raise (Failure ("Non-identifier in output list")) in
    
    (* Remember to setup curv with input variables of gate as Bdec *)
    let rec check_gatedef body (extv, extg) (curv, curg) setl gpmap =
        let lookup_var n local global = try StringMap.find n local
        with Not_found -> try StringMap.find n global
        with Not_found -> raise (Failure ("Undeclared variable: " ^ n)) in

        let lookup_gate n = try StringMap.find n curg
        with Not_found -> try StringMap.find n extg
        with Not_found -> raise (Failure ("Undefined gate: " ^ n)) in

        let check_deref n i local global =
          let var = lookup_var n local global in
          match var with
            Barr (nam, len) -> if i < len && i > -1 then var else raise (Failure
            ("Illegal dereference: " ^ nam ^ ", " ^ n))
          | (Pdec (_, _) | Gdec (_, _) | Bdec _) -> raise (Failure (
              "Illegal dereference: " ^ n ^ ", " ^ string_of_int i)) in

        let rec check_expr e =
            match e with
              Name s -> lookup_var s curv extv; e
            | Index (s, i) -> ignore (check_deref s i curv extv); e
            | Arr el -> List.iter (fun x -> ignore (check_expr x)) el; e
            | Bit _ -> e in

        let rec get_numouts vd =
            match vd with
              Pdec (_, _) -> 1
            | Gdec (n, _) -> (match (lookup_gate n).out with
                Bit _ -> 1
              | Index (_, _) -> 1
              | Arr el -> List.length el
              | Name s -> get_numouts (lookup_var s curv extv))
            | Bdec _ -> 1
            | Barr (_, i) -> i in

        (* Checks types for assignment only *)
        let check_type lhs rhs errf =
            match lhs with
              Bdec _ -> (match rhs with
                Bit _ -> ()
              | Index (_, _) -> ()
              | Name n -> if (get_numouts (lookup_var n curv extv)) == 1 then ()
                else raise errf
              | Arr el -> if (List.length el) == 1 then () else raise errf
              )
            | Barr (_, n) -> (match rhs with
                Arr el -> if (List.length el) == n then () else raise errf
              | Name na -> if (get_numouts (lookup_var na curv extv)) == n then
                  () else raise errf
              | (Bit _ | Index (_, _)) -> raise errf)
            | (Pdec (_, _) | Gdec (_, _)) -> raise errf in

        let check_stmt s setlist =
            let check_no_name n map err =
                if StringMap.mem n map then raise err else n in

            match s with
              Assign (id, expr) -> let rhs = check_expr expr in
                let var = (match id with
                  Symb sy -> lookup_var sy curv StringMap.empty
                | Deref (sy, i) -> check_deref sy i curv StringMap.empty) in
                ignore (check_type var rhs (Failure ("Incompatible assignment: " ^
                string_of_stmt s))); (s :: setlist, curv, curg, gpmap)
            | Vdec vd -> let name = check_no_name (match vd with
                Pdec (_, s) -> s
              | Gdec (g, s) -> ignore (lookup_gate g); s
              | Bdec s -> s
              | Barr (s, _) -> s
              ) curv (Failure ("Cannot redeclare variable: " ^ string_of_stmt s)) in
                (setlist, StringMap.add name vd curv, curg, gpmap)
            | Gdef gd ->
                (try ignore (check_no_name gd.name curg (Failure ("")))
                with failure -> ignore (check_no_name gd.name extg (Failure (
                    "Cannot redeclare gate: " ^ gd.name))));
                let gate = check_gatedef gd.body
                (merge_scopes curv extv, merge_scopes curg extg) (add_inputs
                gd.inp StringMap.empty, StringMap.empty) [] gpmap in
                (setlist, curv, curg, StringMap.add gd.name (
                N.Program (gd.name, gd.inp, gate, build_output_list gd.out)) gpmap)
            | Vset (id, el) -> 
                let var = lookup_var id curv StringMap.empty in
                let err = Failure ("Incorrect number of arguments for gate: "
                ^ id) in
                (match var with
                  Pdec (p, _) ->  (match p with 
                    (And | Or ) -> if List.length el != 2 then raise err
                  | Not -> if List.length el != 1 then raise err )
                | (Bdec _ | Barr (_,_)) -> raise err
                | Gdec (g, _) -> if (List.length (lookup_gate g).inp) != List.length
                el then raise err); (s :: setlist, curv, curg, gpmap)
        in


        match body with
          [] -> Flatten.flat setl curv curg (*Actual interface TBD *)
        | h :: t -> let (sl, cv, cg, gm) = check_stmt h setl in
            check_gatedef t (extv, extg) (cv, cg) sl gm
          
        in
        match name with
          "" -> raise (Failure ("Empty file passed in"))
        | _ -> 
            let gate = check_gatedef body
            (StringMap.empty, StringMap.empty) (add_inputs
            inp StringMap.empty, StringMap.empty) [] StringMap.empty in
            N.Program (name, inp, gate, build_output_list out)
